/****************************************************************************

This file is part of the Common-Demo project on http://qt.gitorious.org.

Copyright (c) 2009 Nokia Corporation and/or its subsidiary(-ies).*
All rights reserved.

Contact:  Nokia Corporation (qt-info@nokia.com)**

You may use this file under the terms of the BSD license as follows:

"Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
* Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
* Neither the name of Nokia Corporation and its Subsidiary(-ies) nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."

****************************************************************************/

#include "items.h"
#include "scene.h"
#include <QWebFrame>
#ifdef DEMO_QML
#include <QmlView>
#endif

class QMLHighlighter : public QSyntaxHighlighter
{
public:
    QMLHighlighter(QTextEdit *edit)
        : QSyntaxHighlighter(edit)
    {
        bold.setBold(true);
        italic.setItalic(true);
    }
    virtual void highlightBlock(const QString &text)
    {
        static QRegExp rx("([A-Za-z0-9._]+:) ([^;]+)");
        int idx = 0;
        while ((idx = rx.indexIn(text, idx)) != -1) {
            setFormat(rx.pos(1), rx.cap(1).size(), Qt::yellow);
            setFormat(rx.pos(2), rx.cap(2).size(), Qt::green);
            idx += rx.matchedLength();
        }

        static QRegExp rx2("([A-Z][A-Za-z0-9]+) \\{");
        idx = 0;
        while ((idx = rx2.indexIn(text, idx)) != -1) {
            setFormat(rx2.pos(1), rx2.cap(1).size(), Qt::red);
            setFormat(rx2.pos(1), rx2.cap(1).size(), bold);
            idx += rx2.matchedLength();
        }

        for (int i=0; i<text.size(); ++i) {
            if (!text.at(i).isLetterOrNumber() && !text.at(i).isSpace()) {
                setFormat(i, 1, italic);
                setFormat(i, 1, Qt::lightGray);
            }
        }
    }
private:
    QFont bold;
    QFont italic;
};

class SourceViewer : public QTabWidget
{
    Q_OBJECT
public:
    SourceViewer(const QString &dirname)
        : QTabWidget()
    {
        QDir dir(dirname);
        dir.setNameFilters(QStringList() << "*.qml" << "*.js");
        QPalette pal = palette();
        pal.setColor(QPalette::Text, Qt::white);
        pal.setColor(QPalette::Foreground, Qt::white);
        pal.setColor(QPalette::Base, Qt::black);
        pal.setColor(QPalette::Background, Qt::black);

        QDirIterator iterator(dir, QDirIterator::Subdirectories);

        while (iterator.hasNext()) {
            QFile f(iterator.next());
            if (f.open(QIODevice::ReadOnly)) {
                QTextStream ts(&f);
                const QString text = ts.readAll();
                QTextEdit *edit = new QTextEdit;
                if (QFileInfo(f).suffix() == "qml")
                    new QMLHighlighter(edit);
                edit->setPalette(pal);
                edit->setReadOnly(true);
                edit->setPlainText(text);
                addTab(edit, f.fileName());
            }
        }
        new QShortcut(Qt::Key_Left, this, SLOT(previous()));
        new QShortcut(Qt::Key_Right, this, SLOT(next()));
        new QShortcut(Qt::Key_Up, this, SLOT(up()));
        new QShortcut(Qt::Key_Down, this, SLOT(down()));
        new QShortcut(Qt::Key_Escape, this, SLOT(close()));
        new QShortcut(Qt::Key_Left + Qt::ControlModifier, this, SLOT(close()));

        setAttribute(Qt::WA_DeleteOnClose);
    }
public slots:
    void next()
    {
        if (count() > 1)
            setCurrentIndex((currentIndex() + 1) % count());
    }

    void previous()
    {
        if (count() > 1) {
            setCurrentIndex((currentIndex() - 1 + count()) % count());
        }
    }

    void up()
    {
        if (QTextEdit *edit = qobject_cast<QTextEdit*>(currentWidget())) {
            edit->verticalScrollBar()->triggerAction(QAbstractSlider::SliderSingleStepSub);
        }
    }

    void down()
    {
        if (QTextEdit *edit = qobject_cast<QTextEdit*>(currentWidget())) {
            edit->verticalScrollBar()->triggerAction(QAbstractSlider::SliderSingleStepAdd);
        }
    }
};

#include "items.moc"

static void setTopLevelWidget(QWidget *w, bool qml)
{
    QStackedWidget *stack = ::stack();
    Q_ASSERT(stack);
    if (qml) {
        w->setParent(stack);
        QRect r(QPoint(), w->sizeHint());
        r.moveCenter(stack->rect().center());
        w->setGeometry(r);
        w->raise();
        w->show();
    } else {
        const int idx = stack->addWidget(w);
        stack->setCurrentIndex(idx);
        QObject::connect(w, SIGNAL(destroyed(QObject*)), stack, SLOT(removeWidget(QObject*)));
    }
    GraphicsScene::instance()->stopDemoMode();
}

void showSource(const QFileInfo &fileInfo)
{
    SourceViewer *sv = new SourceViewer(fileInfo.absoluteFilePath());
    setTopLevelWidget(sv, false);
}


static QPixmap defaultPixmap()
{
    static const QPixmap pixmap = ::loadPixmap(":/resources/image-loading.png");
    return pixmap;
}

static QPixmap pluginPixmap()
{
    static const QPixmap pixmap = ::loadPixmap(":/resources/face-monkey.png");
    return pixmap;
}

static QPixmap folderPixmap()
{
    static const QPixmap pixmap = ::loadPixmap(":/resources/folder.png");
    return pixmap;
}

FileItem::FileItem(const QFileInfo &fileInfo)
    : ViewItem()
{
    d.fileInfo = fileInfo;
    setText(d.fileInfo.fileName());
}

FileItem::FileItem(const QString &file)
    : ViewItem()
{
    d.fileInfo = QFileInfo(file);
    setText(d.fileInfo.fileName());
}

PluginItem::PluginItem(const QFileInfo &fileInfo)
    : FileItem(fileInfo)
{
    d.library = new QLibrary(fileInfo.absoluteFilePath(), this);
    d.library->setLoadHints(d.library->loadHints() | QLibrary::ResolveAllSymbolsHint);
    if (!d.library->load()) {
        qDebug() << d.library->errorString();
        invalidate();
        return;
    }

    Name name = (Name)d.library->resolve("name");
    if (!name) {
        printf("%s %d: if (!name) {\n", __FILE__, __LINE__);
        invalidate();
        return;
    }
    setText(name());
    Icon icon = (Icon)d.library->resolve("icon");
    if (!icon) {
        printf("%s %d: if (!icon) {\n", __FILE__, __LINE__);
        invalidate();
        return;
    }

    d.createWidget = (CreateWidget)d.library->resolve("createWidget");
    if (!d.createWidget) {
        printf("%s %d: if (!d.createWidget) {\n", __FILE__, __LINE__);
        invalidate();
        return;
    }
    const QPixmap pixmap = icon();
    setPixmap(pixmap.isNull() ? pluginPixmap() : pixmap);
}

QVariant ViewItem::itemChange(GraphicsItemChange change, const QVariant &value)
{
    if (change == ItemPositionHasChanged && hasFocus()) {
        emit moved();
    }
    return QGraphicsWidget::itemChange(change, value);
}

QVariant WebItem::itemChange(GraphicsItemChange change, const QVariant &value)
{
    if (change == ItemPositionHasChanged && hasFocus())
        emit moved();
    return QGraphicsWebView::itemChange(change, value);
}

PluginItem::~PluginItem()
{
    delete d.library;
}

bool PluginItem::eventFilter(QObject *o, QEvent *e)
{
    if (e->type() == QEvent::KeyPress) {
        QKeyEvent *keyEvent = static_cast<QKeyEvent*>(e);
        if (false
#ifdef FRIENDLY_KEYS
            || keyEvent->key() == Qt::Key_Escape
#endif
            || (keyEvent->key() == Qt::Key_Left && keyEvent->modifiers() & Qt::ControlModifier)) {
            keyEvent->accept();
            QCoreApplication::instance()->removeEventFilter(this);
            o->deleteLater();
            return true;
        }
    }
    return false;
}

bool PluginItem::exec()
{
    Q_ASSERT(d.library);
    Q_ASSERT(d.createWidget);
    d.createWidget = (CreateWidget)d.library->resolve("createWidget");
    extern void *context();
    QWidget *widget = d.createWidget(context());
    if (!widget) {
        return false;
    }
    widget->setAttribute(Qt::WA_DeleteOnClose);
    QCoreApplication::instance()->installEventFilter(this);
    ::setTopLevelWidget(widget, false);
    return true;
}

#ifdef DEMO_QML
QmlItem::QmlItem(const QFileInfo &info, const QUrl &url)
  : FileItem(info)
{
    d.url = url;
    const QString img = info.absoluteFilePath() + "/dir.png";
    if (QFile::exists(img)) {
        setPixmap(::loadPixmap(img));
    }
}

bool QmlItem::eventFilter(QObject *, QEvent *e)
{
    if (e->type() == QEvent::KeyPress) {
        QKeyEvent *keyEvent = static_cast<QKeyEvent*>(e);
        if (false
#ifdef FRIENDLY_KEYS
            || keyEvent->key() == Qt::Key_Escape
#endif
            || (keyEvent->key() == Qt::Key_Left && keyEvent->modifiers() & Qt::ControlModifier)) {
            keyEvent->accept();
            Q_ASSERT(d.qmlView.data());
            d.qmlView.data()->deleteLater();
            QCoreApplication::instance()->removeEventFilter(this);
            return true;
        }
    } else if (e->type() == QEvent::MouseButtonPress && d.qmlView.data()
               && !d.qmlView.data()->rect().contains(d.qmlView.data()->mapFromGlobal(static_cast<QMouseEvent*>(e)->globalPos()))) {
        d.qmlView.data()->deleteLater();
    }
    return false;
}

class View : public QmlView
{
public:
    View()
    {}

    QRect occupiedArea() const
    {
        QRectF r;
        foreach(QGraphicsItem *item, items()) {
            r |= item->sceneBoundingRect();
        }
        r.moveCenter(viewport()->rect().center());
        return r.toRect();
    }

    void paintEvent(QPaintEvent *e)
    {
        if (buffer.size() != viewport()->size()) {
            buffer = QImage(viewport()->size(), QImage::Format_ARGB32_Premultiplied);
            buffer.fill(0);
        }
        QPainter::setRedirected(viewport(), &buffer);
        QmlView::paintEvent(e);
        QPainter::restoreRedirected(viewport());
#if QT_VERSION >= QT_VERSION_CHECK(4, 7, 0)
        QFileInfo fi(source().toLocalFile());
#else
        QFileInfo fi(url().toLocalFile());
#endif
        QRect r = occupiedArea() & buffer.rect();
        buffer.copy(r).save(fi.absolutePath() + "/dir.png", "PNG");
        QPainter p(viewport());
        p.drawImage(0, 0, buffer);
    }
private:
    QImage buffer;
};

void QmlItem::exec()
{
    Q_ASSERT(!d.qmlView.data());
    Settings settings(fileInfo().absoluteFilePath() + "/settings.conf");
    if (settings.value("captureqmlscreenshots").toBool()) {
        d.qmlView = new View;
    } else {
        d.qmlView = new QmlView;
        d.qmlView.data()->setAttribute(Qt::WA_OpaquePaintEvent);
        d.qmlView.data()->setAttribute(Qt::WA_NoSystemBackground);
    }

    QAction *backAction = new QAction("Back", d.qmlView.data());
    d.qmlView.data()->addAction(backAction);
    backAction->setSoftKeyRole(QAction::PositiveSoftKey);
    connect(backAction, SIGNAL(triggered()), this, SLOT(closeQmlView()));

#if QT_VERSION >= QT_VERSION_CHECK(4, 7, 0)
    d.qmlView.data()->setSource(d.url);
#else
    d.qmlView.data()->setUrl(d.url);
#endif
    d.qmlView.data()->execute();
    if (settings.value("clip", false).toBool()) {
        foreach(QGraphicsItem *item, d.qmlView.data()->items()) {
            item->setFlag(QGraphicsItem::ItemClipsChildrenToShape, true);
        }
    }
    d.qmlView.data()->setAttribute(Qt::WA_DeleteOnClose);
    QCoreApplication::instance()->installEventFilter(this);
    ::setTopLevelWidget(d.qmlView.data(), true);
    QGraphicsView *graphicsView = scene()->views().value(0);
    graphicsView->setEnabled(false);
    connect(d.qmlView.data(), SIGNAL(destroyed(QObject*)), graphicsView, SLOT(reenable()));
    d.qmlView.data()->setFocus();
    connect(QApplication::instance(), SIGNAL(focusChanged(QWidget*,QWidget*)), d.qmlView.data(), SLOT(deleteLater()));
}

void QmlItem::closeQmlView()
{
    d.qmlView.data()->deleteLater();
    QCoreApplication::instance()->removeEventFilter(this);
}

#endif // DEMO_QML

DirectoryItem::DirectoryItem(const QFileInfo &dir)
    : FileItem(dir)
{
    Q_ASSERT(dir.isDir());
    d.dir = dir.absoluteFilePath();
    init();
}

DirectoryItem::DirectoryItem(const QDir &dir)
    : FileItem(dir.absolutePath())
{
    d.dir = dir;
    init();
}

void DirectoryItem::init()
{
    QPixmap pixmap;
    if (d.dir.exists("dir.png")) {
        pixmap = ::loadPixmap(d.dir.absolutePath() + "/dir.png");
    } else {
        QDir copy = d.dir;
        copy.setNameFilters(QStringList() << "*.png");
        QDirIterator it(copy);
        if (it.hasNext()) {
            QString fn = d.dir.absolutePath();
            fn.append('/');
            fn.append(it.next());
            pixmap = ::loadPixmap(fn);
        }
    }
    setPixmap(pixmap.isNull() ? folderPixmap() : pixmap);
}

ImageItem::ImageItem(const QFileInfo &fileInfo)
    : FileItem(fileInfo)
{
    setPixmap(defaultPixmap());
}

ImageItem::ImageItem(const QPixmap &pixmap, const QFileInfo &fileInfo)
    : FileItem(fileInfo)
{
    setPixmap(pixmap);
}

void BackButton::mousePressEvent(QGraphicsSceneMouseEvent *event)
{
    event->accept();
}

void BackButton::mouseReleaseEvent(QGraphicsSceneMouseEvent *event)
{
    emit clicked();
}

void BackButton::paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *)
{
    painter->fillRect(option->rect, Qt::green);
}

#ifdef DEMO_GL_FBO
void ViewItem::paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *)
{
    painter->beginNativePainting();
    GLWidget *glWidget = qobject_cast<GLWidget *>(scene()->views().first()->viewport());
    // save state changed by painter->beginNativePainting
    glWidget->saveGLState();
    if (!d.pixmap.isNull()) {
        glEnable(GL_TEXTURE_2D);
        if (!d.textureId)
            d.textureId = glWidget->bindTexture(d.pixmap, GL_TEXTURE_2D, GL_RGBA,
                                              QGLContext::LinearFilteringBindOption | QGLContext::InvertedYBindOption);
        else
            glBindTexture(GL_TEXTURE_2D, d.textureId);
    } else {
        glBindTexture(GL_TEXTURE_2D, 0);
    }

    glEnable(GL_CULL_FACE);
    gluLookAt(-2, -2, 5, 0.0, 0.0, 0.0, 0.0, 10, 0.0);

    glTranslatef(option->rect.center().x(), option->rect.center().y(), 0);
    glWidget->drawCube(option->rect.size().width()/2);
    glWidget->restoreGLState();
    painter->endNativePainting();
}
#elif defined(DEMO_GLES_FBO)
void ViewItem::paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *)
{
    painter->beginNativePainting();
    GLWidget *glWidget = qobject_cast<GLWidget *>(scene()->views().first()->viewport());
    // bind texture
    if (!d.pixmap.isNull()) {
        glEnable(GL_TEXTURE_2D);
        if (!d.textureId)
            d.textureId = glWidget->bindTexture(d.pixmap, GL_TEXTURE_2D, GL_RGBA,
                                                QGLContext::LinearFilteringBindOption | QGLContext::InvertedYBindOption);
        else
            glBindTexture(GL_TEXTURE_2D, d.textureId);
    } else {
        glBindTexture(GL_TEXTURE_2D, 0);
    }
    glEnable(GL_CULL_FACE);

    QMatrix4x4 matrix;
    matrix.ortho(0, glWidget->width(), glWidget->height(), 0, -option->rect.width()/2, 50);
    matrix.translate(painter->transform().dx(), painter->transform().dy());
    matrix.translate(option->rect.center().x(), option->rect.center().y(), 0);
    matrix.lookAt(QVector3D(-2, -2, 5), QVector3D(0.0, 0.0, 0.0), QVector3D(0.0, 10, 0.0));

//    glEnable(GL_BLEND);							// Enable Blending (Otherwise The Reflected Object Wont Show)
//    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);			// Blending Based On Source Alpha And 1 Minus Dest Alpha

    glWidget->drawCube(option->rect.size().width()/2, matrix);

//    glDisable(GL_BLEND);
    glDisable(GL_CULL_FACE);
    painter->endNativePainting();
}
#else
void ViewItem::paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *)
{
    if (d.fillBackground) {
        painter->fillRect(option->rect, QColor(0, 0, 0, 200));
    }

    if (d.polygon) {
        QPolygonF from(QRectF(option->rect));
        from.resize(4);
        QTransform transform;
        QTransform::quadToQuad(from, *d.polygon, transform);
        painter->save();
        painter->setTransform(transform);
    }

    const QBrush brush = option->palette.background();
    if (!d.pixmap.isNull()) {
        const QRect pixmapRect = d.pixmap.rect();
        QSize s = pixmapRect.size();
        s.scale(option->rect.size(), Qt::KeepAspectRatio);
        QRect r(QPoint(0, 0), s);
        r.moveCenter(option->rect.center());
        painter->drawPixmap(r, d.pixmap, pixmapRect);
    }
    const QString t = text();
    if (d.textVisible && !t.isEmpty()) {
        QFont font;
        font.setPixelSize(qMax<int>(8, size().height() / 20));
        painter->setFont(font);
        enum { TextMargin = 10 };
        painter->setPen(option->state & QStyle::State_HasFocus ? Qt::yellow : Qt::white);
        painter->drawText(option->rect.adjusted(TextMargin, 0, 0, 0), Qt::AlignTop|Qt::AlignLeft, t);
    }
    if (d.polygon) {
        painter->restore();
    }
}
#endif

WebItem::WebItem(const QFileInfo &fileInfo)
{
    setFlag(ItemIsFocusable);
    d.fileInfo = fileInfo;
    QUrl url;
    QFile file(fileInfo.absoluteFilePath());
    if (file.open(QIODevice::ReadOnly)) {
        url = QTextStream(&file).readLine();
    }
    init(url);
    if (!QCoreApplication::instance()->property("noplugins").toBool())
        settings()->setAttribute(QWebSettings::PluginsEnabled, true);
    addActions();
}

WebItem::WebItem(const QUrl &url)
{
    init(url);
    addActions();
}

void WebItem::init(const QUrl &url)
{
    d.loading = false;
    setSizePolicy(QSizePolicy::Preferred, QSizePolicy::Preferred);
    connect(this, SIGNAL(loadFinished(bool)), this, SLOT(onLoadFinished()));
    connect(this, SIGNAL(loadProgress(int)), this, SLOT(onLoadProgress()));
    load(url);
}

void WebItem::addActions()
{
//    pageAction(QWebPage::Reload)->setShortcut(QKeySequence::Refresh);
//    addAction(pageAction(QWebPage::Reload));

    pageAction(QWebPage::Back)->setShortcut(QKeySequence::Back);
    addAction(pageAction(QWebPage::Back));

    pageAction(QWebPage::Forward)->setShortcut(QKeySequence::Forward);
    addAction(pageAction(QWebPage::Forward));
}

void WebItem::paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *widget)
{
    if (d.loading) {
        painter->fillRect(option->rect, Qt::white);
    }
    QGraphicsWebView::paint(painter, option, widget);
}

void WebItem::resizeEvent(QGraphicsSceneResizeEvent *event)
{
    page()->setViewportSize(size().toSize());
    QGraphicsWebView::resizeEvent(event);
}

ViewItem::ViewItem()
    : QGraphicsWidget()
{
    d.polygon = 0;
    d.fillBackground = false;
    d.textVisible = true;
    d.yRotation = d.xRotation = 0;
    setFlag(ItemIsFocusable);
    setSizePolicy(QSizePolicy::Preferred, QSizePolicy::Preferred);
#if defined(DEMO_GL_FBO) || defined(DEMO_GLES_FBO)
    d.textureId = 0;
#endif
}

ViewItem::~ViewItem()
{
#if defined(DEMO_GL_FBO) || defined(DEMO_GLES_FBO)
    QGraphicsView *view = scene()->views().value(0);
    if (d.textureId && view) {
        GLWidget *glWidget = qobject_cast<GLWidget *>(view->viewport());
        if (glWidget)
            glWidget->deleteTexture(d.textureId);
    }
#endif
}

void ViewItem::setYRotation(qreal rotation)
{
    d.yRotation = rotation;
    QTransform transform;
    const QPointF origin = rect().center();
    transform.translate(origin.x(), origin.y());
    transform.rotate(rotation, Qt::YAxis);
    transform.translate(-origin.x(), -origin.y());
    setTransform(transform);
}

void ViewItem::setXRotation(qreal rotation)
{
    d.xRotation = rotation;
    QTransform transform;
    const QPointF origin = rect().center();
    transform.translate(origin.x(), origin.y());
    transform.rotate(rotation, Qt::XAxis);
    transform.translate(-origin.x(), -origin.y());
    setTransform(transform);
}
ProgressBarItem::ProgressBarItem(int itemCount)
{
    d.itemCount = itemCount;
}

void ProgressBarItem::paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *)
{
    if (d.itemCount == 0)
        return;
    qreal total = 0;
    foreach(int value, d.progress) {
        total += value;
    }
    qreal w = (total / qreal(d.itemCount * 100)) * option->rect.width();
    static QPixmap gradient;
    if (gradient.isNull()) {
        gradient = QPixmap(100, 100);
        QLinearGradient g(0, 0, 0, gradient.height());
        g.setColorAt(0.0, Qt::black);
        g.setColorAt(.5, Qt::gray);
        g.setColorAt(1.0, Qt::black);
        QPainter(&gradient).fillRect(gradient.rect(), g);
    }

    const QRectF r(option->rect.x(), option->rect.y(), w, option->rect.height());
    painter->fillRect(r, gradient);
}

void ProgressBarItem::setValue(int progress)
{
    d.progress[sender()] = progress;
    qreal total = 0;
    foreach(int value, d.progress) {
        total += value;
    }
    if (total == d.itemCount * 100) {
        deleteLater();
    } else {
        update();
    }
}

void WebItem::keyPressEvent(QKeyEvent *e)
{
    if (e->modifiers() == Qt::ControlModifier && page()) {
        enum { Add = 20 };
        int add = Add;
        switch (e->key()) {
        case Qt::Key_B:
        case Qt::Key_Left:
            add = -add;
        case Qt::Key_F:
        case Qt::Key_Right:
            page()->mainFrame()->setScrollBarValue(Qt::Horizontal, page()->mainFrame()->scrollBarValue(Qt::Horizontal) + add);
            return;
        case Qt::Key_P:
        case Qt::Key_Up:
            add = -add;
        case Qt::Key_N:
        case Qt::Key_Down:
            page()->mainFrame()->setScrollBarValue(Qt::Vertical, page()->mainFrame()->scrollBarValue(Qt::Vertical) + add);
            return;
        }
    }
    QGraphicsWebView::keyPressEvent(e);
}
void ViewItem::mousePressEvent(QGraphicsSceneMouseEvent *event)
{
    if (event->button() == Qt::LeftButton) {
        GraphicsScene *theScene = static_cast<GraphicsScene*>(scene());
        theScene->setCurrentItem(this);
        theScene->centerOn(this);
        event->accept();
    } else {
        QGraphicsWidget::mousePressEvent(event);
    }
}

void ViewItem::mouseReleaseEvent(QGraphicsSceneMouseEvent *event)
{
    if (event->button() == Qt::LeftButton) {
        GraphicsScene *theScene = static_cast<GraphicsScene*>(scene());
        theScene->activate(this);
        event->accept();
    } else {
        QGraphicsWidget::mousePressEvent(event);
    }
}


void TextItem::paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *widget)
{
    if (d.background.style() != Qt::NoBrush) {
        painter->fillRect(option->rect, d.background);
    }
    QGraphicsSimpleTextItem::paint(painter, option, widget);
}
QPolygonF ViewItem::polygon() const
{
    if (d.polygon)
        return *d.polygon;
    QPolygonF poly(geometry());
    poly.resize(4);
    return poly;
}

void ViewItem::setPolygon(const QPolygonF &pp)
{
    if (pp.isEmpty()) {
        delete d.polygon;
        d.polygon = 0;
        const QRectF oldGeometry = property("oldGeometry").toRectF();
        setProperty("oldGeometry", QVariant());
        setGeometry(oldGeometry);
    } else {
        if (!d.polygon) {
            setProperty("oldGeometry", geometry());
            d.polygon = new QPolygonF(pp);
        } else {
            *d.polygon = pp;
        }
        setGeometry(pp.boundingRect().adjusted(-1, -1, 1, 1));
    }
}

QPainterPath ViewItem::shape() const
{
    QPainterPath path;
    if (d.polygon) {
        path.addPolygon(*d.polygon);
    } else {
        path.addRect(boundingRect());
    }
    path.closeSubpath();
    return path;
}

