/****************************************************************************

This file is part of the Common-Demo project on http://qt.gitorious.org.

Copyright (c) 2009 Nokia Corporation and/or its subsidiary(-ies).*
All rights reserved.

Contact:  Nokia Corporation (qt-info@nokia.com)**

You may use this file under the terms of the BSD license as follows:

"Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
* Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
* Neither the name of Nokia Corporation and its Subsidiary(-ies) nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."

****************************************************************************/

#ifndef ITEMS_H
#define ITEMS_H

#include <QtGui>
#include "scene.h"
#include <QGraphicsWebView>

#ifdef DEMO_GL_FBO
#include "glwidget.h"
#elif defined(DEMO_GLES_FBO)
#include "gleswidget.h"
#endif

enum ItemType { FileType = QGraphicsItem::UserType + 1, DirectoryType, ImageType, WebPageType, VideoType, PluginType, QmlType, ReflectionType };

static inline QStackedWidget *stack()
{
    foreach(QWidget *w, QApplication::topLevelWidgets()) {
        if (QStackedWidget *stack = qobject_cast<QStackedWidget*>(w)) {
            return stack;
        }
    }
    return 0;
}

class BackButton : public QGraphicsWidget
{
Q_OBJECT
public:
    void paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *widget = 0);

protected:
    void mousePressEvent(QGraphicsSceneMouseEvent *event);
    void mouseReleaseEvent(QGraphicsSceneMouseEvent *event);

Q_SIGNALS:
    void clicked();
};

Q_DECLARE_METATYPE(QFileInfo);
class ViewItem : public QGraphicsWidget
{
    Q_OBJECT
    Q_PROPERTY(int itemType READ itemType)
    Q_PROPERTY(qreal xRotation READ xRotation WRITE setXRotation)
    Q_PROPERTY(qreal yRotation READ yRotation WRITE setYRotation)
    Q_PROPERTY(QString text READ text WRITE setText)
    Q_PROPERTY(bool textVisible READ isTextVisible WRITE setTextVisible)
    Q_PROPERTY(bool fillBackground READ fillBackground WRITE setFillBackground)
    Q_PROPERTY(QPolygonF polygon READ polygon WRITE setPolygon)
public:
    ViewItem();
    ~ViewItem();

    void setPixmap(const QPixmap &pixmap) {
        d.pixmap = pixmap;
        update();
#if defined(DEMO_GL_FBO) || defined(DEMO_GLES_FBO)
        Q_ASSERT(!d.textureId); // we don't support changing pixmap yet
#endif
    }
    QPixmap pixmap() const { return d.pixmap; }

    virtual void paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *widget = 0);
    QString text() const { return d.text; }
    void setText(const QString &text) { d.text = text; update(); }
    virtual int itemType() const = 0;

    qreal xRotation() const { return d.xRotation; }
    qreal yRotation() const { return d.yRotation; }
    void setYRotation(qreal rotation);
    void setXRotation(qreal rotation);

    void resetTransform() /*  overrridden */ { QGraphicsWidget::resetTransform(); d.xRotation = d.yRotation = 0.0; }
    bool isTextVisible() const { return d.textVisible; }
    void setTextVisible(bool textVisible) { d.textVisible = textVisible; update(); }
    virtual QVariant itemChange(GraphicsItemChange change, const QVariant &value);
    void setFillBackground(bool fill) { d.fillBackground = fill; update(); }
    bool fillBackground() const { return d.fillBackground; }

    QPolygonF polygon() const;
    void setPolygon(const QPolygonF &pp);
    QPainterPath shape() const;
    void mousePressEvent(QGraphicsSceneMouseEvent *event);
    void mouseReleaseEvent(QGraphicsSceneMouseEvent *event);

signals:
    void moved();
protected:
    struct Data {
        qreal xRotation, yRotation;
        QPixmap pixmap;
        QString text;
        bool textVisible;
        bool fillBackground;
        QPolygonF *polygon;
#if defined(DEMO_GL_FBO) || defined(DEMO_GLES_FBO)
        GLint textureId;
#endif
    } d;
};

class ReflectionItem : public ViewItem
{
    Q_OBJECT
        public:
    ReflectionItem(const QPixmap &pixmap)
        : ViewItem()
    {
        setPixmap(pixmap);
    }

    virtual int itemType() const { return ReflectionType; }
};

class TextItem : public QObject, public QGraphicsSimpleTextItem
{
    Q_OBJECT
    Q_PROPERTY(qreal opacity READ opacity WRITE setOpacity)
    Q_PROPERTY(QBrush background READ background WRITE setBackground)
public:
    TextItem(const QString &string = QString()) : QGraphicsSimpleTextItem(string) { setCacheMode(ItemCoordinateCache); }
    virtual void paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *widget = 0);
    QBrush background() const { return d.background; }
    void setBackground(const QBrush &background) { d.background = background; update(); }
private:
    struct Data {
        QBrush background;
    } d;
};

class FileItem : public ViewItem
{
    Q_OBJECT
    Q_PROPERTY(QFileInfo fileInfo READ fileInfo)
public:
    FileItem(const QFileInfo &fileInfo);
    FileItem(const QString &file);
    virtual int itemType() const { return FileType; }
    QFileInfo fileInfo() const { return d.fileInfo; }
    virtual QString text() const { return fileInfo().fileName(); }
private:
    struct Data {
        QFileInfo fileInfo;
    } d;
};

#ifdef DEMO_QML
class QmlView;
class QmlItem : public FileItem
{
    Q_OBJECT
    Q_PROPERTY(QUrl url READ url)
public:
    QmlItem(const QFileInfo &info, const QUrl &url);
    virtual int itemType() const { return QmlType; }
    QUrl url() const { return d.url; }
    void exec();
    bool eventFilter(QObject *o, QEvent *e);

protected Q_SLOTS:
    void closeQmlView();

private:
    struct Data {
        QUrl url;
        QWeakPointer<QmlView> qmlView;
    } d;
};
#endif

typedef QWidget *(*CreateWidget)(void *);
typedef QPixmap (*Icon)();
typedef QString (*Name)();
class PluginItem : public FileItem
{
    Q_OBJECT
public:
    PluginItem(const QFileInfo &fileInfo);
    ~PluginItem();
    bool eventFilter(QObject *o, QEvent *e);
    bool exec();
    virtual int itemType() const { return PluginType; }
    bool isValid() const { return d.library; }
    void invalidate() { delete d.library; d.library = 0; }
private:
    struct Data {
        QLibrary *library;
        CreateWidget createWidget;
    } d;
};

Q_DECLARE_METATYPE(QDir);
class DirectoryItem : public FileItem
{
    Q_OBJECT
    Q_PROPERTY(QDir dir READ dir)
public:
    DirectoryItem(const QFileInfo &dir);
    DirectoryItem(const QDir &dir);
    virtual int itemType() const { return DirectoryType; }
    QDir dir() const { return d.dir; }
private:
    void init();
    struct Data {
        QDir dir;
    } d;
};

class DotDotItem : public DirectoryItem
{
public:
    DotDotItem(const QFileInfo &fileInfo) : DirectoryItem(fileInfo) {}
    DotDotItem(const QDir &dir) : DirectoryItem(dir) {}
    virtual QString text() const { return QLatin1String(".."); }
};

class WebItem : public QGraphicsWebView
{
    Q_OBJECT
    Q_PROPERTY(int itemType READ itemType)
    Q_PROPERTY(QFileInfo fileInfo READ fileInfo)
public:
    WebItem(const QFileInfo &fileInfo);
    WebItem(const QUrl &url);
    virtual void paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *widget = 0);
    virtual void resizeEvent(QGraphicsSceneResizeEvent *event);
    virtual int itemType() const { return WebPageType; }
    QFileInfo fileInfo() const { return d.fileInfo; }
    virtual QVariant itemChange(GraphicsItemChange change, const QVariant &value);
    virtual void keyPressEvent(QKeyEvent *e);
signals:
    void moved();
private slots:
    void onLoadFinished() { d.loading = false; update(); }
    void onLoadProgress() { if (!d.loading) { d.loading = true; update(); } }
private:
    void init(const QUrl &url);
    void addActions();
    struct Data {
        bool loading;
        QFileInfo fileInfo;
    } d;
};

class ProgressBarItem : public QGraphicsWidget
{
    Q_OBJECT
public:
    ProgressBarItem(int itemCount);
    virtual void paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *widget = 0);
public slots:
    void setValue(int progress);
private:
    struct Data {
        QHash<QObject*, int> progress;
        int itemCount;
    } d;
};

class ImageItem : public FileItem
{
    Q_OBJECT
public:
    ImageItem(const QPixmap &pixmap, const QFileInfo &fileInfo);
    ImageItem(const QFileInfo &fileInfo);
    virtual int itemType() const { return ImageType; }
public slots:
    void showSmallPixmap()
    {
        if (d.bigPixmap.isNull())
            return;
        setPixmap(d.smallPixmap);
        d.smallPixmap = QPixmap();
        d.bigPixmap = QPixmap();
    }
    void showEnlarged()
    {
        QString fileName = fileInfo().absoluteFilePath();
        fileName.replace("_thumbnail", "_full");
        const QSize s = scene()->views().first()->viewport()->size();
        d.bigPixmap = ::loadPixmap(fileName, s);
        if (!d.bigPixmap.isNull()) {
            d.smallPixmap = pixmap();
            setPixmap(d.bigPixmap);
        }
    }
private:
    struct Data {
        QPixmap smallPixmap;
        QPixmap bigPixmap;
    } d;
};

class SelectorItem : public QGraphicsWidget
{
    Q_OBJECT
public:
    SelectorItem()
    {}

    void select(QGraphicsWidget *widget)
    {
        if (widget != item.data()) {
            if (item)
                disconnect(item.data(), 0, this, 0);
            setVisible(widget);
            item = widget;
            if (item) {
                adjustGeometry();
                connect(item.data(), SIGNAL(moved()), this, SLOT(adjustGeometry()));
            }
        } else if (widget) {
            adjustGeometry();
        }
    }
    virtual void paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget * = 0)
    {
        QPen pen(Qt::yellow, 2);
        pen.setCosmetic(true);
        painter->setPen(pen);
        painter->setBrush(Qt::NoBrush);
        painter->drawRect(option->rect.adjusted(2, 2, -3, -3));
    }

public slots:
    void adjustGeometry()
    {
        if (item) {
            enum { Margin = 4 };
            setGeometry(item.data()->geometry().adjusted(-Margin, -Margin, Margin, Margin));
        }
    }

private:
    QWeakPointer<QGraphicsWidget> item;
};


#endif
