/****************************************************************************

This file is part of the Common-Demo project on http://qt.gitorious.org.

Copyright (c) 2009 Nokia Corporation and/or its subsidiary(-ies).*
All rights reserved.

Contact:  Nokia Corporation (qt-info@nokia.com)**

You may use this file under the terms of the BSD license as follows:

"Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
* Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
* Neither the name of Nokia Corporation and its Subsidiary(-ies) nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
THIS SOFTWARE IS  PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."

****************************************************************************/

#include <QtGui>
#ifdef DEMO_GLVIEWPORT
#include <QtOpenGL>
#endif
#ifdef DEMO_QML
#include <QmlView>
#endif

#include "scene.h"
#include "settings.h"
#include "view.h"

static QVariant interpolateQPolygonF(const QPolygonF &from, const QPolygonF &to, qreal progress)
{
    if (qFuzzyIsNull(progress)) {
        return qVariantFromValue(from);
    } else if (qFuzzyCompare(progress, (qreal)1.0)) {
        return qVariantFromValue(to);
    }
    const int count = to.size();
    Q_ASSERT(count == from.size());
    QPolygonF ret;
    ret.resize(count);

    for (int i=0; i<count; ++i) {
        ret[i] = (from.at(i) + ((to.at(i) - from.at(i)) * progress));
    }
    return qVariantFromValue(ret);
}

class StackedWidget : public QStackedWidget
{
    Q_OBJECT
    Q_PROPERTY(bool allowDemoMode READ allowDemoMode)
public:
    StackedWidget()
    {}

    void showEvent(QShowEvent *e)
    {
        if (Settings().value("activatewindow").toBool()) {
            activateWindow();
            raise();
        }
        QStackedWidget::showEvent(e);
    }

    bool allowDemoMode() const
    {
#ifdef DEMO_QML
        return count() == 1 && !qFindChild<QmlView*>(this);
#else
        return count() == 1;
#endif
    }
public slots:
    void removeWidget(QObject *o)
    {
        QStackedWidget::removeWidget(static_cast<QWidget*>(o));
    }
};

#ifdef DEMO_GLVIEWPORT
QGLContext *ctx = 0;
void *context()
{
    return ctx;
}
#else
void *context()
{
    return 0;
}
#endif

static QString findPictureFolder()
{
    QDir dir("/data/demo/");
    dir.setNameFilters(QStringList() << "*Pictures*");
    dir.setFilter(QDir::NoDotAndDotDot|QDir::Dirs);
    foreach(const QString &name, dir.entryList()) {
        const QString dirName = dir.absoluteFilePath(name);
        Settings settings;
        if (!settings.value("hidden", false).toBool()) {
            return dirName;
        }
    }
    return QString();
}

int main(int argc, char **argv)
{
    srand(QTime(0, 0, 0).msecsTo(QTime::currentTime()));
    QPixmapCache::setCacheLimit(20*1024);
    qRegisterMetaType<QDir>("QDir");
    qRegisterMetaType<QList<QKeyEvent*> >("QList<QKeyEvent*>");
    qRegisterMetaType<QFileInfo>("QFileInfo");
    qRegisterMetaType<QPolygonF>("QPolygonF");
    qRegisterAnimationInterpolator<QPolygonF>(interpolateQPolygonF);

    QApplication::setOrganizationName("Nokia");
    QApplication::setOrganizationDomain("qt.nokia.com");
    QApplication::setApplicationName("CES Demo");
#ifdef GL_GRAPHICSSYSTEM
    QApplication::setGraphicsSystem("opengl");
#elif defined Q_WS_X11
    QApplication::setGraphicsSystem("raster");
#endif
    QApplication a(argc, argv, QApplication::GuiServer);
    QDir::setCurrent("C:/");


    for (int i=1; i<argc - 1; ++i) {
        const QString arg = QString::fromLatin1(argv[i]);
        if (arg == "--config" || arg == "-config") {
            a.setProperty("configs", (a.property("configs").toStringList() << QString::fromLatin1(argv[i + 1])));
            argv[i] = 0;
            argv[++i] = 0;
        }
    }

    Settings settings;
    QRect geometry = QApplication::desktop()->screenGeometry();
    geometry.moveLeft(settings.value("geometry/x", geometry.x()).toInt());
    geometry.moveTop(settings.value("geometry/y", geometry.y()).toInt());
    geometry.setWidth(settings.value("geometry/width", geometry.width()).toInt());
    geometry.setHeight(settings.value("geometry/height", geometry.height()).toInt());

    const bool fullScreen = settings.value("fullscreen", false).toBool();
    const bool background = settings.value("background", true).toBool();
    a.setProperty("fitInView", settings.value("fitInView", true).toBool());
    a.setProperty("itemSize", settings.value("itemSize", 300).toInt());
    a.setProperty("maxItems", settings.value("maxItems", INT_MAX).toInt());
    a.setProperty("idleTimer", settings.value("idletimer", 30000).toInt());
    a.setProperty("replayFile", settings.value("replayFile").toString());
    a.setProperty("nowebpreload", settings.value("nowebpreload", true));
    const int cursorBlinkPeriod = settings.value("cursorFlashTime", -1).toInt();
    if (cursorBlinkPeriod >= 0)
        a.setCursorFlashTime(cursorBlinkPeriod);

    const int imageThrottle = settings.value("imageThrottle", -1).toInt();
    if (imageThrottle > 0)
        a.setProperty("--image-throttle", imageThrottle);
    a.setProperty("nowebpreload", settings.value("nowebpreload", false).toBool());
    a.setProperty("noplugins", settings.value("noplugins", false).toBool());

    for (int i=1; i<argc; ++i) {
        if (!argv[i])
            continue;

        const QString arg = QString::fromLatin1(argv[i]);
        if (arg == "-record" || arg == "--record") {
            QString file;
            if (i + 1 < argc) {
                file = argv[++i];
            } else {
                printf("%s %d: } else {\n", __FILE__, __LINE__);
                return 1;
            }
            a.setProperty("record", file);
        }
    }
    if (background) {
        QDir dir(::findPictureFolder());
        dir.setFilter(QDir::Files);
        dir.setNameFilters(QStringList() << "*full.png");
        const QStringList list = dir.entryList();
        if (!list.isEmpty()) {
            const int i = rand() % list.size();
            if (!list.value(i).isEmpty())
                a.setProperty("backgroundImage", dir.absolutePath() + "/" + list.value(i));
        }

    }
    StackedWidget stack;
    QShortcut *s = new QShortcut(QKeySequence(Qt::CTRL + Qt::Key_C), &stack, SLOT(close()));
    s->setContext(Qt::ApplicationShortcut);

    GraphicsView *view = new GraphicsView;
    stack.addWidget(view);

    if (fullScreen) {
        stack.showFullScreen();
    } else {
        stack.setWindowFlags(stack.windowFlags() | Qt::FramelessWindowHint);
        stack.setGeometry(geometry);
        stack.show();
    }
    return a.exec();
}

#include "main.moc"
