/****************************************************************************

This file is part of the Common-Demo project on http://qt.gitorious.org.

Copyright (c) 2009 Nokia Corporation and/or its subsidiary(-ies).*
All rights reserved.

Contact:  Nokia Corporation (qt-info@nokia.com)**

You may use this file under the terms of the BSD license as follows:

"Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
* Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
* Neither the name of Nokia Corporation and its Subsidiary(-ies) nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."

****************************************************************************/

#ifndef SCENE_H
#define SCENE_H

#include <QtGui>
#include "settings.h"
#include "loadpixmap.h"

struct Event {
    int key;
    int modifiers;
    int msecs;
};

class ImageLoadingThread : public QThread
{
    Q_OBJECT
public:
    ImageLoadingThread(int throttle, const QSizeF &size, QObject *parent = 0);
    void run();
    void setImages(const QList<QPair<QString, void*> > &images);
    void abort();
    bool isSuspended() const;
    void suspend();
    void resume();
signals:
    void imageLoaded(const QImage &image, void *userData);
private:
    struct Data {
        QList<QPair<QString, void*> > images;
        QMutex mutex;
        int imageSize;
        int throttle;
        bool suspended;
        mutable QMutex suspensionMutex;
    } d;
};

Q_DECLARE_METATYPE(QPolygonF);

class QMenu;
class TextItem;
class ViewItem;
class WebItem;
class GraphicsView;
class ScriptInterface;
class ImageItem;
class SelectorItem;
class ProgressBarItem;
class CoverFlowManager;
class BackButton;
class GraphicsScene : public QGraphicsScene
{
    Q_OBJECT
    Q_PROPERTY(QSizeF itemSize READ itemSize)
    Q_ENUMS(LayoutMode)
public:
    GraphicsScene(GraphicsView *view);
    static GraphicsScene *instance();
    ~GraphicsScene();
    QSizeF itemSize() const;
    void keyPressEvent(QKeyEvent *e);
    void drawBackground(QPainter *p, const QRectF &rr);
    virtual void drawForeground(QPainter *painter, const QRectF &rect);
    enum Direction {
        None,
        Left,
        Right,
        Up,
        Down
    };

    QGraphicsWidget *findItem(Direction direction) const;
    QGraphicsWidget *currentItem() const { return static_cast<QGraphicsWidget*>(focusItem()); }
    void setCurrentItem(QGraphicsWidget *newCurrent);
    void timerEvent(QTimerEvent *e);
    int duration() const;
    QEasingCurve curve() const;
    void activate(QGraphicsWidget *widget);
    enum LayoutMode {
        Collapse = 0,
        Grid,
        Shuffle,
        Ellipse,
        Outline,
        Diagonal,
        Diagonal2,
        ZoomedAndIsolated ,
#ifdef ROTATIONS_ENABLED
        CoverFlow,
        RotateX,
        RotateY,
        RotateZ,
#endif
        LayoutModeCount
    };

    void setDirectory(const QDir &dir, LayoutMode mode = Grid, const QString &newCurrent = QString());

    QRectF visibleSceneRect() const;
    QRectF availableScreenRect() const;
    void centerOn(QGraphicsWidget *item);
    void resetLayout() { setLayout(d.currentLayoutMode); }
    QPointF gridStartPos() const;
    void start();
signals:
    void directoryClicked(const QDir &dir);
public slots:
#ifdef ROTATIONS_ENABLED
    void resetRotations();
#endif
    void activateCurrentItem();
    void onLayoutFinished();
    void setLayout(LayoutMode mode);
    void onImageLoaded(const QImage &image, void *ptr);
    void stopDemoMode();
    void startDemoMode();
    void restartDemoMode();
    void updateDemoModePosition();
    void onWebItemsLoaded();
    void showMenu();
    void goParentDirectory();
private:
    void computeSceneRect();
    int layoutModeForKey(int key, Qt::KeyboardModifiers modifiers) const;
    void restoreCurrentItem();

    QAbstractAnimation *createShuffle() const;
    QAbstractAnimation *createEllipse() const;
    QAbstractAnimation *createGrid() const;
    QAbstractAnimation *createOutline() const;
    QAbstractAnimation *createDiagonal() const;
    QAbstractAnimation *createDiagonal2() const;
    QAbstractAnimation *createCollapse() const;
    QAbstractAnimation *createZoomedAndIsolated() const;
#ifdef ROTATIONS_ENABLED
    QAbstractAnimation *createRotate(Qt::Axis axis) const;
#endif

    void queryUrlDialog();

    struct Data {
        int demoModeDuration;
        TextItem *demoModeItem;
        QList<QGraphicsWidget*> items;
        QBasicTimer demoTimer;
        bool demoMode;
        ScriptInterface *scriptInterface;
        QString root;
        QWeakPointer<ImageLoadingThread> imageLoadingThread;
        QBasicTimer suspendThreadTimer;
        QWeakPointer<QPropertyAnimation> currentViewScaleAnimation;
        QSet<ImageItem*> pendingImageItems;
        QPropertyAnimation *verticalScrollBarAnimation;
        qreal nextZOrder;
        QWeakPointer<QAbstractAnimation> lastAnimation;
        LayoutMode currentLayoutMode;
        GraphicsView *view;
        int frames;
        QTime started;
        TextItem *fpsItem;
        QWeakPointer<QAbstractAnimation> fpsItemFadeAnimation;
        SelectorItem *selector;
        QString pendingDirectory;
        QFile *record;
        QTime recordTime;
        QList<Event> replayEvents;
        int replayEventIdx;
        QDir *dotDot;
        QMenu *menu;
        QDir current;
        QHash<QString, WebItem*> cachedWebItems;
        QWeakPointer<ProgressBarItem> progressItem;
        QWeakPointer<BackButton> backButton;
        QBasicTimer idleTimer;
        bool sendingDemoModeEvent;
        bool directoryTextVisible;
        bool onlySelectedTextVisible;
        QBitArray supportedTransitions;

        QPixmap backgroundPixmap;
        QString backgroundFileName;
        QGraphicsView::ViewportUpdateMode pendingViewportUpdateMode;
        CoverFlowManager *coverFlowManager;
        bool hasNonViewItems;
    } d;
    friend class CoverFlowManager;
    static GraphicsScene *inst;
};

#endif
