/****************************************************************************

This file is part of the Common-Demo project on http://qt.gitorious.org.

Copyright (c) 2009 Nokia Corporation and/or its subsidiary(-ies).*
All rights reserved.

Contact:  Nokia Corporation (qt-info@nokia.com)**

You may use this file under the terms of the BSD license as follows:

"Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
* Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
* Neither the name of Nokia Corporation and its Subsidiary(-ies) nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."

****************************************************************************/

#ifndef VIEW_H
#define VIEW_H

#include <QtGui>
#include <QWidget>
#include <QShortcut>
#include <QKeySequence>
#include <QShowEvent>
#include <QTimer>
#include <QDir>
#include <QPixmap>
#include "scene.h"
#ifdef DEMO_GLVIEWPORT
#include <QtOpenGL>
#endif
#ifdef DEMO_GL_FBO
#include "glwidget.h"
#endif
#ifdef DEMO_GLES_FBO
#include "gleswidget.h"
#endif

class GraphicsView : public QGraphicsView
{
    Q_OBJECT
    Q_PROPERTY(qreal zoom READ zoom WRITE setZoom)
public:
    GraphicsView(QWidget *parent = 0)
        : QGraphicsView(parent)
    {
        viewport()->setAutoFillBackground(false);
        d.scene = 0;
        setFrameStyle(0);
#ifdef DEMO_GL
        setViewportUpdateMode(FullViewportUpdate);
#ifdef DEMO_GLVIEWPORT
        QGLWidget *widget = new QGLWidget(this);
        extern QGLContext *ctx;
        ctx = const_cast<QGLContext*>(widget->context());
        setViewport(widget);
#elif defined(DEMO_GL_FBO) || defined(DEMO_GLES_FBO)
        setViewport(new GLWidget(this));
#endif
#else
        setViewportUpdateMode(BoundingRectViewportUpdate);
#endif
        setAlignment(Qt::AlignTop|Qt::AlignLeft);
        setOptimizationFlag(DontClipPainter, true);
        setOptimizationFlag(DontSavePainterState, true);
        setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
        setVerticalScrollBarPolicy(Qt::ScrollBarAlwaysOff);

        QShortcut *s = new QShortcut(QKeySequence(Qt::CTRL + Qt::Key_S), this, SLOT(executeScript()));
        s->setContext(Qt::ApplicationShortcut);
        scale(0.5, 0.5);
    }

    void keyPressEvent(QKeyEvent *e)
    {
        switch (e->key()) {
        case Qt::Key_Equal:
        case Qt::Key_Plus:
            if (Settings().value("noscale").toBool())
                return;
                scale(1.1, 1.1);
            break;
        case Qt::Key_Minus:
            if (Settings().value("noscale").toBool())
                return;
            scale(0.9, 0.9);
            break;
#if defined(DEMO_GL_FBO) || defined(DEMO_GLES_FBO)
        case Qt::Key_AsciiTilde:
        case Qt::Key_Dollar:
            if (GLWidget *vp = qobject_cast<GLWidget *>(viewport())) {
                vp->nextSurface();
            }
            break;
        case Qt::Key_A:
        case Qt::Key_D:
            if (!(e->modifiers() & Qt::ControlModifier)) {
                QGraphicsView::keyPressEvent(e);
                return;
            }
            if (GLWidget *vp = qobject_cast<GLWidget *>(viewport())) {
                e->key() == Qt::Key_A ? vp->moveCameraLeft() : vp->moveCameraRight();
            }
            break;

        case Qt::Key_W:
        case Qt::Key_Z:
            if (!(e->modifiers() & Qt::ControlModifier)) {
                QGraphicsView::keyPressEvent(e);
                return;
            }
            if (GLWidget *vp = qobject_cast<GLWidget *>(viewport())) {
                e->key() == Qt::Key_W ? vp->moveCameraForward() : vp->moveCameraBackward();
            }
            break;
#endif
        case Qt::Key_R:
            if (e->modifiers() & (Qt::AltModifier|Qt::ControlModifier)) {
                setTransform(QTransform());
                break;
            }
            // fallthrough
        default:
            QGraphicsView::keyPressEvent(e);
            return;
        }
        d.scene->resetLayout();
    }

    void mousePressEvent(QMouseEvent *event)
    {
        d.pressPoint = event->pos();
        d.originalSceneRect = sceneRect();
        QGraphicsView::mousePressEvent(event);
    }

    void mouseMoveEvent(QMouseEvent *event)
    {
        QRectF rect = sceneRect();
        rect.moveTop(d.originalSceneRect.y() - (1/transform().m11())*(event->pos().y() - d.pressPoint.y()));
        setSceneRect(rect);
        QGraphicsView::mouseMoveEvent(event);
    }

    void mouseReleaseEvent(QMouseEvent *event)
    {
        if (qAbs(event->pos().y() - d.pressPoint.y()) > 20) {
            event->ignore();
        } else {
            QGraphicsView::mouseReleaseEvent(event);
        }
    }


    void zoomTo(QGraphicsWidget *item)
    {
        fitInView(item->geometry(), Qt::KeepAspectRatio);
    }

    void closeEvent(QCloseEvent *e)
    {
        QGraphicsView::closeEvent(e);
    }

    void showEvent(QShowEvent *e)
    {
        static bool first = true;
        if (first) {
            QTimer::singleShot(100, this, SLOT(start()));
            first = false;
        }
        setFocus();
        QGraphicsView::showEvent(e);
    }

    QRectF view() const
    {
        return mapToScene(viewport()->rect()).boundingRect();
    }

    qreal zoom() const
    {
        // ### this relies on there only being scales/transforms set
        // ### on the view and that both x any y are scaled the same
        return transform().m11();
    }

    void setZoom(qreal zoom)
    {
        QTransform transform;
        transform.scale(zoom, zoom);
        setTransform(transform);
    }

public slots:
    void reenable()
    {
        setEnabled(true);
        setFocus();
    }
    void start()
    {
        d.scene = new GraphicsScene(this);
        setScene(d.scene);
#if defined(DEMO_GL_FBO) || defined(DEMO_GLES_FBO)
        GLWidget *vp = qobject_cast<GLWidget*>(viewport());
        Q_ASSERT(vp != 0);
        vp->setScene(d.scene);
#endif
        d.scene->start();
    }

    void executeScript()
    {
        QFile script(QApplication::applicationDirPath() + "/script.sh");
        if (!script.exists())
            return;
        QProcess *process = new QProcess(this);
        QString program = "sh";
        QStringList args;
        args << "script.sh";
        process->execute(program, args);
        connect(process, SIGNAL(finished(int, QProcess::ExitStatus)), process, SLOT(deleteLater()));
    }

    void zoomIn()
    {
        scale(1.1, 1.1);
    }

    void zoomOut()
    {
        scale(0.9, 0.9);
    }

protected:
    void paintEvent(QPaintEvent *event)
    {
#if defined(DEMO_GL_FBO) || defined(DEMO_GLES_FBO)
        GLWidget *vp = qobject_cast<GLWidget *>(viewport());
        Q_ASSERT(vp);
        vp->draw();
#else
        QGraphicsView::paintEvent(event);
#endif
    }

private:
    struct Data {
        GraphicsScene *scene;
        QPoint pressPoint;
        QRectF originalSceneRect;
#if defined(DEMO_GL_FBO) || defined(DEMO_GLES_FBO)
#endif
    } d;
};

#endif
