/****************************************************************************

This file is part of the Common-Demo project on http://qt.gitorious.org.

Copyright (c) 2009 Nokia Corporation and/or its subsidiary(-ies).*
All rights reserved.

Contact:  Nokia Corporation (qt-info@nokia.com)**

You may use this file under the terms of the BSD license as follows:

"Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
* Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer. 
* Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution. 
* Neither the name of Nokia Corporation and its Subsidiary(-ies) nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission. 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."

****************************************************************************/

#include <QApplication>
#include <QtGui>
#include <QWebView>

class PageContainer : public QScrollArea
{
public:
    PageContainer(QWidget *p)
        : QScrollArea(p), currentIndex(0)
    {
        setWidgetResizable(true);
        setSizePolicy(QSizePolicy::MinimumExpanding, QSizePolicy::MinimumExpanding);
        sceneWidget = new QWidget;
        layout = new QHBoxLayout(sceneWidget);
        layout->setMargin(0);
        layout->setSpacing(0);
        setWidget(sceneWidget);
        translateViewport = new QPropertyAnimation(horizontalScrollBar(), "value", this);
        translateViewport->setDuration(600);
        translateViewport->setEasingCurve(QEasingCurve::InOutQuad);
        setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
    }
    void keyPressEvent(QKeyEvent *)
    {}

    QWebView *addPage(const QUrl &address)
    {
        QWebView *newView = new QWebView;
        layout->addWidget(newView);
        newView->setUrl(address);
        pages << newView;
        return newView;
    }

    void previousPage()
    {
        changePage(currentIndex-1, true);
    }

    void nextPage()
    {
        changePage(currentIndex+1, true);
    }

    void changePage(int newIndex, bool animated = false)
    {
        if (pages.count() <=0) {
            return;
        }
        translateViewport->stop();

        currentIndex = qBound(0,newIndex,pages.count()-1);

        QWebView *selectedPage = pages.at(currentIndex);
        if (!dynamic_cast<QLineEdit*>(QApplication::focusWidget())) {
            selectedPage->setFocus(Qt::OtherFocusReason);
        }

        const int hmax = sceneWidget->width(); // horizontalScrollBar()->maximum(); why does this not work?
        int endValue = (newIndex*hmax)/pages.count();

        if (animated) {
            translateViewport->setStartValue(horizontalScrollBar()->value());
            translateViewport->setEndValue(endValue);
            translateViewport->start();
        } else {
            horizontalScrollBar()->setValue(endValue);
        }
    }

    QWebView *currentPage() const
    {
        return pages.value(currentIndex);
    }


protected:
    void resizeEvent(QResizeEvent *event)
    {
        foreach(QWebView *page, pages) {
            page->setFixedSize(event->size());
        }
        QScrollArea::resizeEvent(event);
    }

private:
    QHBoxLayout *layout;
    QWidget *sceneWidget;
    QList<QWebView*> pages;
    int currentIndex;
    QPropertyAnimation *translateViewport;
};

class Widget : public QWidget
{
    Q_OBJECT
public:
    Widget()
        : QWidget(0), addressBar(new QLineEdit(this)), pageContainer(new PageContainer(this))
    {
        addressBar->installEventFilter(this);

        new QShortcut(QKeySequence(Qt::CTRL | Qt::Key_L), addressBar, SLOT(setFocus()));

        QVBoxLayout *vboxLayout = new QVBoxLayout(this);
        vboxLayout->setSpacing(0);
        vboxLayout->setContentsMargins(0,0,0,0);
        vboxLayout->addWidget(addressBar);
        vboxLayout->addWidget(pageContainer);

        forgeConnections();

        pageContainer->installEventFilter(this);
        const char *urls[] = {  "http://www.google.com", "http://www.cnn.com", "http://www.onion.com", 0 };

        for (int i = 0; urls[i]; ++i) {
            QWebView *webView = pageContainer->addPage(QUrl(QString::fromLatin1(urls[i])));
            webView->installEventFilter(this);
        }
    }

    void forgeConnections()
    {
        connect(addressBar,
                SIGNAL(returnPressed()),
                SLOT(openPage()));
    }

    bool eventFilter(QObject *watched, QEvent *event)
    {
        if (event->type() == QEvent::ShortcutOverride || event->type() == QEvent::KeyPress) {
            QKeyEvent *keyEvent = static_cast<QKeyEvent*>(event);
            if (keyEvent->modifiers() & Qt::ShiftModifier) {
                switch (keyEvent->key()) {
                case Qt::Key_Left:
                    if (event->type() == QEvent::KeyPress) {
                        pageContainer->previousPage();
                    }
                    event->accept();
                    return true;
                case Qt::Key_Right:
                    if (event->type() == QEvent::KeyPress) {
                        pageContainer->nextPage();
                    }
                    event->accept();
                    return true;
                default:
                    break;
                }
            } else if (keyEvent->modifiers() && Qt::ControlModifier) {
                if (keyEvent->key() == Qt::Key_Left) {
                    deleteLater();
                    return true;
                }
            }
        }
        return QWidget::eventFilter(watched, event);
    }


public slots:
    void openPage()
    {
        QUrl url(QUrl::fromUserInput(addressBar->text()));
        pageContainer->currentPage()->setUrl(url);
        addressBar->setText(url.toString());
        addressBar->selectAll();
    }

private:
    QLineEdit *addressBar;
    PageContainer *pageContainer;
};

#include "../plugin.h"

DEFINE_PLUGIN_FUNCS(minibrowser, "Browser", Widget);

#include "main.moc"
